\ COPYRIGHT 1993 DUTCH FORTH WORKSHOP - FILENAME : FLTEST.FRT #2
\ LANGUAGE    : tForth
\ PROJECT     : Testing of the floating point package
\ DESCRIPTION : Tabulates/plots deviations of the FEXP etc.
\ CATEGORY    : Test
\ AUTHOR      : Albert van der Horst
\ LAST CHANGE : ANSI 961206 AH VALUES need initialisation
\ LAST CHANGE : 930419 AH Use MAX-FLOAT
\ LAST CHANGE : 930418 AH ABORT possibility
\ LAST CHANGE : 930309 AH WORD instead of postponed S"
\ LAST CHANGE : 930306 AH EvaluateFunc added
\ LAST CHANGE : 930306 AH Longer marker at 0
\ LAST CHANGE : 930214 AH changed for simpler header
\ CREATED     : 930212 AH

        NEEDS -graphics
        NEEDS -strings

        REVISION -fltest " Plotting Function Values  Version 0.3t "

PRIVATES

DEFER function                                      \ F: <r1> --- <r2>

\ Auxiliary, give a possibility for a break
: BREAK     BREAK? IF TEXT ABORT THEN ;

0FVALUE start PRIVATE            \ start of interval
0FVALUE width PRIVATE            \ width of interval

0VALUE ABSOLUTE
1 TO ABSOLUTE
\ 2^MantissaBits is the machine precision, often called epsilon
S" DOUBLE-PRECISION" ENVIRONMENT? 0= [IF]
  .( Fatal: no DOUBLE-PRECISION, is this tForth?)
[ELSE]
  [IF] 52 [ELSE] 23 [THEN] CONSTANT MantissaBits PRIVATE
[THEN]
2e0 FLN MantissaBits S>F F* FEXP FCONSTANT EPSSCALE PRIVATE

\ >Eps takes a floating point number representing a deviation as a
\ second parameter of the floating point stack. The fist parameter serves
\ for scaling, to determine the machine precision attainable.
\ It returns the number of Epilons that the deviation represents.
\ So 0 is exact, 1 is the least possible deviation etc.
\ In short i = f2/f1 scaled in machine precision units.
: >Eps    ( f: f2 f1 -- f )

       FDUP F0= IF
         FDROP FDROP 0e0
       ELSE
         FSPLIT FSWAP
         FSPLIT FDROP           \ f: f1'    d: m1 m2
         - MantissaBits +       \ f1' * 2 ^tos
         BEGIN                  \ f1' * 2 ^tos  is the deviations in eps
            DUP                 \ and a loop invariant
         WHILE
            FDUP F+             \ Double
            1-
         REPEAT
         DROP
       ENDIF
; PRIVATE

\ The table contains
\  1 INTEGER  : start of interval in points
\  2 INTEGER  : end of interval in points
\  3 INTEGER  : number of points per unit
\  4 FLOAT    : value of function at start of interval
\  5 FLOAT    : idem one point farther
\  ...
\  x FLOAT    : value of function at end of interval

\ Return the integer bounds of a table
\ Appropriate for a DO .. LOOP
: Bounds        ( funtable -- high+1.low )
    DUP @ SWAP CELL+ @ 1+ SWAP
; PRIVATE

\ Return an XY pair at index i in the table
: GetXY   ( funtable i --     f:  -- fx fy )
   2DUP S>F 2 CELLS + @ S>F F/        \ Got fx
   OVER @ -                     \ Offset in table
   SWAP 3 CELLS +               \ Table address
   []FLOAT F@                   \ Got fy
; PRIVATE

: XY>Eps
      FSWAP function FSWAP
      FOVER F-                  \ function, delta
      ABSOLUTE IF
         FSWAP FDROP EPSSCALE F*
      ELSE
         >Eps
      ENDIF
; PRIVATE

: ShowFuncDeviations2    ( Table-address -- . )
    DUP Bounds DO
      DUP I GetXY XY>Eps
      CR ." eps " F.
    LOOP
    DROP
;


\ Search for the maximum error between the approximated and the real function.
0FVALUE funmax     PRIVATE         0VALUE maxix     PRIVATE
0FVALUE funmin     PRIVATE         0VALUE minix     PRIVATE
0FVALUE funsma     PRIVATE         0VALUE smallest  PRIVATE
0FVALUE funlar     PRIVATE         0VALUE largest   PRIVATE

0FVALUE MAX-FLOAT
S" MAX-FLOAT" ENVIRONMENT? DROP TO MAX-FLOAT
MAX-FLOAT F. MAX-FLOAT FNEGATE F.

: MAXMIN   ( table address )
         LOCAL Table
         MAX-FLOAT FNEGATE TO funmax  1 TO maxix
         MAX-FLOAT TO funmin  1 TO minix
         MAX-FLOAT TO funsma  1 TO smallest
         0e0 TO funlar  1 TO largest

        Table Bounds DO
             Table I GetXY XY>Eps
             FDUP  funmax F> IF  FDUP TO funmax  I TO maxix    ENDIF
             FDUP  funmin F< IF  FDUP TO funmin  I TO minix    ENDIF
             FABS
             FDUP  funsma F< IF  FDUP TO funsma  I TO smallest ENDIF
             FDUP  funlar F> IF  FDUP TO funlar  I TO largest  ENDIF
             FDROP
        LOOP
; PRIVATE


: ShowStatistics ( <table address> -- )
     MAXMIN
   CR ." funmax  " funmax   F. ."  maxix    " maxix      .
   CR ." funmin  " funmin   F. ."  minix    " minix      .
   CR ." funsma  " funsma   F. ."  smallest " smallest   .
   CR ." funlar  " funlar   F. ."  largest  " largest    .
; PRIVATE

: InitStart
   DUP 2 CELLS + @ LOCAL Scale
   DUP @ S>F Scale S>F F/ TO start
   DUP CELL+ @ SWAP @ - S>F Scale S>F F/ TO width
; PRIVATE

\ Draw a border around the picture
: BORDER
        0    0    Xmax 0    lightred DRAW-LINE
        Xmax 0    Xmax Ymax lightred DRAW-LINE
        Xmax Ymax 0    Ymax lightred DRAW-LINE
        0    Ymax 0    0    lightred DRAW-LINE
; PRIVATE

\ Draw a vertical line to mark the unit positions, only if
\ X turns to be an integer.
: ?DrawMarker        ( X -- )
    .5e0 FLOCAL LEN                     \ Default length of the marker
    FDUP F>S S>F FOVER 0e0 F~ IF        \ Exactly equal to an integer
       FDUP F0= IF 1e0 TO LEN ENDIF     \ Larger at X=0

       FDUP TO PenX LEN FNEGATE TO PenY
       FDUP LEN lightred DRAW-SCALED-LINE
    THEN
    FDROP
; PRIVATE

?DEF DEBUG [IF]

\ Test BORDER and ?DrawMarker
\ Expect a frame round the window plus markers with the origin
\ to the left and below the middle
: Test?Drawframe ( -- )
   GRAPHICS
   BORDER
   -10e -5e0 +15e0 +7e0 SET-GWINDOW
   16 -10 DO I S>F ?DrawMarker LOOP
   KEY TEXT
;
[THEN]

: ShowGraphics     ( table address -- )
    LOCAL TABLE
        TABLE MAXMIN
        TABLE InitStart
        start funmin 1e0 F-   start width F+ funmax 1e0 F+  SET-GWINDOW
        BORDER
        TABLE Bounds DO
           TABLE I GetXY
           FOVER TO PenX
           FOVER ?DrawMarker
           XY>Eps TO PenY
           white PLOT-POINT
        LOOP
;



STRING FUNC$  PRIVATE  20 NEW FUNC$  \ Going to contain FEXP FLN etc.
STRING INCL$  PRIVATE 100 NEW INCL$ \ INCLUDE command
STRING DEFER$ PRIVATE  20 NEW DEFER$ \ Deferred function
STRING TABLE-PATH$    100 NEW TABLE-PATH$ \ Where the tables are
S" " TO TABLE-PATH$                     \ Current directory

\ Evaluate a floating point function. From the function a statistic of the
\ deviations is shown, plus a plot. In the directory TABLE-PATH$
\ it expects a file that, if interpreted, generates a table with true
\ values such as described above.
\ After the plot this table is forgotten.
\ Interprets the word following in the input stream: the name of
\ a Forth floating point function. The Boolean indicates that we
\ want a plot of the absolute not relative deviation.
\ The EVALUATE's are necessary because at compile time the
\ FUNCTABLE does not exist yet.
: EvaluateFunc        ( flag -- )
    TO ABSOLUTE
    BL WORD COUNT TO FUNC$      \ Get word from input stream
    S" INCLUDE " TO INCL$       \ Generate file name
    TABLE-PATH$ +TO INCL$
    FUNC$       +TO INCL$
    S" .FTB"    +TO INCL$
    CR ." Building table of " FUNC$ TYPE ."
    INCL$ EVALUATE              \ Include that file
    S" ' " TO    DEFER$         \ Generate defer command
    FUNC$ +TO    DEFER$
    DEFER$ EVALUATE IS function \ Install the function wanted in `fltest'

    TEXT
    CR ." Presenting statistics of " FUNC$ TYPE
    CR ." Deviation is " ABSOLUTE IF ." absolute" ELSE ." relative" THEN
    S" FUNCTABLE" EVALUATE  ShowStatistics
    CR ." Hit a key to go on."
    BREAK

    CR ." Presenting graphics of " FUNC$ TYPE ." , hit a key to go on, "
    CR ." also hit a key to go on after graphics "
    BREAK
    GRAPHICS
    S" FUNCTABLE " EVALUATE ShowGraphics
    BREAK
    TEXT

    S" FORGET FUNCTABLE" EVALUATE
;

?DEF DEBUG [IF]
STRING COCO$ 100 NEW COCO$ \ COPY command

\ Function to compare to: constant 0
: TEST FDROP 0e0 ;

\ Test the EvaluateFunc with a table that should generate a plot with
\ 25 ascending dots, two pro marker in the interval -2.5 to 10 .
\ Markers are miniscule, because the error is colossal.
\ Expect statistics to give errors of 1/EPSSCALE
\ Assumes the echo and copy commands are imported from the OS.
: TestEvaluateFunc

  \ Generate an FTB file for TEST
  S" echo CREATE FUNCTABLE \ : 25 points >  TEST.FTB" EVALUATE
  S" echo -12 , 12 , 2 ,                          >> TEST.FTB" EVALUATE
  S" echo -5e0 F, -4e0 F, -3e0 F, -2e0 F, -1e0 F, >> TEST.FTB" EVALUATE
  S" echo  0e0 F,  1e0 F,  2e0 F,  3e0 F,  4e0 F, >> TEST.FTB" EVALUATE
  S" echo  5e0 F,  6e0 F,  7e0 F,  8e0 F,  9e0 F, >> TEST.FTB" EVALUATE
  S" echo 10e0 F, 11e0 F, 12e0 F, 13e0 F, 14e0 F, >> TEST.FTB" EVALUATE
  S" echo 15e0 F, 16e0 F, 17e0 F, 18e0 F, 19e0 F, >> TEST.FTB" EVALUATE

  \ Copy it to the official directory
  S" copy TEST.FTB " TO COCO$   TABLE-PATH$ +TO COCO$
  COCO$ EVALUATE

  \ Do it
  S" 1 EvaluateFunc TEST" EVALUATE
;
[THEN]

DEPRIVE
